#ifndef TWN_UTIL_H
#define TWN_UTIL_H

#include "twn_types.h"
#include "twn_engine_api.h"

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <math.h>


#ifndef M_PI
#define M_PI 3.14159265358979323846264338327950288   /**< pi */
#endif

/* multiply by these to convert degrees <---> radians */
#define DEG2RAD (M_PI / 180)
#define RAD2DEG (180 / M_PI)

#ifndef TWN_NOT_C

    TWN_API void *cmalloc(size_t size);
    TWN_API void *crealloc(void *ptr, size_t size);
    TWN_API void *ccalloc(size_t num, size_t size);

    TWN_API void log_info(const char *restrict format, ...);
    TWN_API void log_critical(const char *restrict format, ...);
    TWN_API void log_warn(const char *restrict format, ...);

    /* saves all texture atlases as BMP files in the write directory */
    TWN_API void textures_dump_atlases(void);

    /* returns true if str ends with suffix */
    TWN_API bool strends(const char *str, const char *suffix);

    /* TODO: this is why generics were invented. sorry, i'm tired today */
    TWN_API double clamp(double d, double min, double max);
    TWN_API float clampf(float f, float min, float max);
    TWN_API int clampi(int i, int min, int max);

    /* sets buf_out to a pointer to a byte buffer which must be freed. */
    /* returns the size of this buffer. */
    TWN_API int64_t file_to_bytes(const char *path, unsigned char **buf_out);

    /* returns a pointer to a string which must be freed */
    TWN_API char *file_to_str(const char *path);

    /* returns true if the file exists in the filesystem */
    TWN_API bool file_exists(const char *path);

#endif /* TWN_NOT_C */

/* calculates the overlap of two rectangles */
TWN_API Recti overlap_recti(const Recti a, const Recti b);
TWN_API Rect overlap_rect(const Rect a, const Rect b);

/* returns true if two rectangles are intersecting */
TWN_API bool intersect_recti(const Recti a, const Recti b);
TWN_API bool intersect_rect(const Rect a, const Rect b);

/* TODO: generics and specials (see m_vec2_from() for an example)*/
TWN_API Recti to_recti(Rect rect);
TWN_API Rect to_rect(Recti rect);

TWN_API Vec2i center_recti(Recti rect);
TWN_API Vec2 center_rect(Rect rect);

/* decrements an integer value, stopping at 0 */
/* meant for tick-based timers in game logic */
/*
 * example:
 * tick_timer(&player->jump_air_timer);
 */
TWN_API int32_t timer_tick_frames(int32_t frames_left);

/* decrements a floating point second-based timer, stopping at 0.0f */
/* meant for poll based real time logic in game logic */
/* note that it should be decremented only on the next tick after its creation */
TWN_API float timer_tick_seconds(float seconds_left);

TWN_API struct timer_elapse_frames_result {
    bool elapsed; int32_t frames_left;
} timer_elapse_frames(int32_t frames_left, int32_t interval);

TWN_API struct timer_elapse_seconds_result {
    bool elapsed; float seconds_left;
} timer_elapse_seconds(float seconds_left, float interval);

#endif
